(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: lrope.ml,v 1.9 2014/01/21 09:21:17 fclement Exp $ *)

(* Linear ropes, optimized for big buffers handling *)

(* invariant: len is never zero *)
type lrope =
  (* buf, off, len, total_len *)
  | S of (string * int * int * int * (lrope Lazy.t))
  | Nil
;;
type t = lrope;;

let length = function
  | S (_, _, _, l, _) -> l
  | Nil -> 0
;;

let empty = Nil;;

let string_append lr s =
  let l = String.length s in
  if l = 0 then lr else
  S (s, 0, l, length lr + l, lazy lr)
;;

let rec rev_list_append lr l =
  match l with
  | s :: t -> rev_list_append (string_append lr s) t
  | [] -> lr
;;

let list_append lr l = rev_list_append lr (List.rev l);;

let rec index_from_to s f t c =
  if f >= t then None else
  if String.unsafe_get s f = c then Some f else index_from_to s (succ f) t c
;;

let index lr c =
  let rec in_index lr d =
    match lr with
    | S (s, off, len, _, lazy lr) ->
      begin match index_from_to s off (off + len) c with
      | None -> in_index lr (d + len)
      | Some i -> Some (d + i)
      end
    | Nil -> None in
  in_index lr 0
;;

let take_n lr n =
  if n >= Sys.max_string_length || n > length lr
    then invalid_arg "take_n" else
  let rec in_take out off rem lr =
    match lr with
    | S (s, offs, lens, _, lazy lr) ->
      let l = min rem lens in
      String.blit s offs out off l;
      if rem = l then ()
      else in_take out (off + l) (rem - l) lr
    | Nil -> () in
  let s = String.create n in in_take s 0 n lr; s
;;

let skip_n lr n =
  if n >= Sys.max_string_length || n > length lr
    then invalid_arg "take_n" else
  let rec in_skip rem lr =
    match lr with
    | S (s, offs, lens, l, lazy lr) ->
      if rem < lens
      then S (s, offs + rem, lens - rem, l - rem, lazy lr)
      else in_skip (rem - lens) lr
    | Nil -> Nil in
  in_skip n lr
;;

let get_c lr =
  match lr with
  | S (s, off, _, _, _) -> s.[off]
  | Nil -> invalid_arg "get_c"
;;

let rec concat lr0 lr1 =
  match lr0 with
  | Nil -> lr1
  | S (s, off, len, l, lazy lr) ->
    S (s, off, len, l + length lr1, lazy (concat lr lr1))
;;
